var camera = {
    canvas: null,
    ctx: null,
    steps: [],
    halfStep: 0,
    dist: [],
    sectionCount: Game.projectionWidth / Game.sectWidth,
    cast: function () {
      this.ctx.clearRect(0,0,Game.projectionWidth,Game.projectionHeight);
      this.drawBackground();
      this.dist[this.sectionCount] = 0;
      for (let i = 0; i < this.sectionCount; i++) {
        this.dist[i] = this.trackRay(this.steps[i] + player.rotation, i, player.x, player.y);
        this.dist[this.sectionCount] = Math.max(this.dist[i], this.dist[this.sectionCount]);
      }
      this.renderSprites();
    },
    trackRay: function(angle, strip, ex, ey){
        angle = only360(angle);
        let isRight = angle > Game.twoPi * 0.75 || angle < Game.twoPi * 0.25;
        let isUp = angle > Math.PI;
        let distance = 0;
        let hit = {x: 0, y: 0};
        let playerX = ex / Game.gameScale;
        let playerY = ey / Game.gameScale;
        let wallType = 0;
        let textureX = 0;
        let slope = Math.sin(angle)/Math.cos(angle);
        let xChange = isRight ? 1 : -1; // xChange is in block steps
        let yChange = xChange * slope;
        // check left or right for ray collision with vertical wall
        let x = isRight ? Math.ceil(playerX) : Math.floor(playerX);
        let y = playerY + (x - playerX) * slope;
        let mapWidth = Game.map.getWidth();
        let mapHeight = Game.map.getHeight();
        while(x >= 0 && x < mapWidth && y >= 0 && y < mapHeight){
          let wallX = Math.floor(x + (isRight ? 0 : -1));
          let wallY = Math.floor(y);
          if(Game.map.getGrid(wallX, wallY)){ // we have a hit so calc the diatance
            distance = Math.hypot(x * Game.gameScale - ex, 
                                  y * Game.gameScale - ey);
            hit.x = x;
            hit.y = y;
            textureX = isRight? 1 - y % 1 : y % 1;
            wallType = Game.map.getGrid(wallX, wallY); // for later
            break;
          }
          x += xChange;
          y += yChange;
        }
        // now check up and down the grid
        slope = Math.cos(angle) / Math.sin(angle);
        yChange = isUp ? -1 : 1;
        xChange = yChange * slope;
        y = isUp ? Math.floor(playerY) : Math.ceil(playerY);
        x = playerX + (y - playerY) * slope;
        while(x >= 0 && x < mapWidth && y >= 0 && y < mapHeight){
          wallY = Math.floor(y + (isUp ? -1 : 0));
          wallX = Math.floor(x);
          if(Game.map.getGrid(wallX, wallY)){
            let newDistance = Math.hypot(x * Game.gameScale - ex,
                                          y * Game.gameScale - ey);
            if(!distance || newDistance < distance){
              distance = newDistance;
              hit.x = x;
              hit.y = y;
              textureX = isUp ? 1 - x % 1 : x % 1;
              wallType = Game.map.getGrid(wallX, wallY);
            }
            break;
          }
          y += yChange;
          x += xChange;
        }
        if(strip > -1){
          Game.map.drawRay(player, hit);
          // correct distance for fish eye distortion
          let noFish = distance * Math.cos(player.rotation - angle);
          this.renderStrip(strip, wallType, noFish, textureX);
        }
        return distance;
    },
    renderStrip: function(strip, wallType, distance, textureX) {
      let height = Game.playerDistance / distance;
      let width = Math.round(Game.sectWidth * (1 / height)); //correct lateral distortion
      if(width < 1) {width = 1;}
      height = Math.round(height * Game.gameScale);
      textureX = Math.round(textureX * Game.gameScale);
      if (textureX > Game.gameScale - width) { //needs correcting
          textureX = Game.gameScale - width;
      }
      let top = Math.round((Game.projectionHeight - height) / 2);
      this.ctx.drawImage(Game.walls[wallType - 1], textureX, 0, width, Game.gameScale, strip * Game.sectWidth, top, Game.sectWidth, height);
    },
    drawBackground: function() {
      this.ctx.beginPath();
      this.ctx.fillStyle = Level.bgTopGrad;
      this.ctx.rect(0,0,this.canvas.width, this.canvas.height);
      this.ctx.closePath();
      this.ctx.fill();
      this.ctx.beginPath();
      this.ctx.fillStyle = Level.bgFloorGrad;
      this.ctx.rect(0,this.canvas.height/2,this.canvas.width, this.canvas.height);
      this.ctx.closePath();
      this.ctx.fill();    
    },
    renderSprites: function() {
      let dx = 0, dy = 0, da = 0, dd = 0, step = 0;
      for(let i = 0; i < sprites.length; i++){
        dx = sprites[i].x - player.x;
        dy = sprites[i].y - player.y;
        sprites[i].distance = Math.hypot(dx, dy);
        sprites[i].angle = this.frontAngle(dx, dy, 0); // true angle to x/y
      }
      // sort in descending order of distances so we can draw from furthest to nearest
      if(sprites.length > 1){
        sprites.sort(function(a,b){return b.distance - a.distance;});
      }
      // idea is that sprite sits in circle with diameter of sprite width
      // and centre at point x/y
      // we are looking for intersections from player to diameter of circle
      // perpendicular to line from player
      for(let i = 0; i < sprites.length; i++){
        if(sprites[i].distance > this.dist[this.sectionCount] || !sprites[i].visible){continue;} // too far away
        let dims = sprites[i].getDims(); // get height and width
        if(sprites[i].viewedFrom){ // has it got the method?
          sprites[i].viewedFrom(player.x, player.y); // 3D sprite or NPC
        }
      let lPoint = findPointOnCircle(sprites[i].x, sprites[i].y, dims.width/2, sprites[i].angle - Math.PI/2);
      let rPoint = findPointOnCircle(sprites[i].x, sprites[i].y, dims.width/2, sprites[i].angle + Math.PI/2);
      // the sprite is on the line from lPoint to rPoint we need to see which part of the line is visible
      // to the player in the FOV
      let xStart = null, xEnd = null;
      xStep = (rPoint.x - lPoint.x) / dims.width;
      yStep = (rPoint.y - lPoint.y) / dims.width;
      for(let s = 0; s < dims.width; s++){
          dx = (lPoint.x + s * xStep) - player.x; // x position of pixel on sprite image
          dy = (lPoint.y + s * yStep) - player.y;
          da = this.frontAngle(dx, dy, player.rotation); // relative angle to point on sprite
          dd = Math.hypot(dx, dy); // distance to point on Sprite
          step = this.getAnglePos(da,dd); // is angle in FOV?
          if(step > -1){
              xStart = {x: s, step: step};
              break;
          }
      }
      if(xStart){
          for(let s = dims.width - 1; s > 0; s--){
              dx = (lPoint.x + s * xStep) - player.x;
              dy = (lPoint.y + s * yStep) - player.y;
              da = this.frontAngle(dx, dy, player.rotation); // relative angle
              dd = Math.hypot(dx, dy);
              step = this.getAnglePos(da,dd); // in FOV ?
              if(step > -1){
                  xEnd = {x: s, step: step};
                  break;
              }
          }
      }
      if(xEnd){
          let distance = sprites[i].distance * (Math.cos(sprites[i].angle - player.rotation)); // correct the distance
          let height = Game.playerDistance / distance;
          let wallHeight = Math.round(height * Game.gameScale); // height of notional wall at same position
          let spriteHeight = Math.round(height * dims.height);
          let width = Math.abs(xEnd.x - xStart.x);
          let spriteWidth = Math.round(width * height);
          let top = Math.round((Game.projectionHeight - wallHeight) / 2);
          top = top + wallHeight - spriteHeight;
          let xPos = xStart.step * Game.sectWidth;
          // watch the brackets below (N.B. Sprite3D is also an instanceof Sprite)
          if((sprites[i] instanceof Sprite3D) || (sprites[i] instanceof NPC)){
            this.ctx.drawImage(sprites[i].canvas, xStart.x, 0, width, dims.height, xPos, top, spriteWidth, spriteHeight);
            if(sprites[i] instanceof NPC){sprites[i].state |= NPC.states.seen;}
          } else {
            this.ctx.drawImage(sprites[i].img, xStart.x, 0, width, dims.height, xPos, top, spriteWidth, spriteHeight);
          }
      }
    }
  },
  getAnglePos: function(da, dd){
    if(da >= this.steps[0]){
        // where is relative angle da in FOV?
        for(let a = 0; a < this.sectionCount-1; a++){
            if(da >= this.steps[a] - this.halfStep && da < this.steps[a] + this.halfStep){
                if(dd <= this.dist[a]) {
                    return a;
                }
            }
        }
    }
    return -1; // angle not in FOV
  },
    init: function(canvId){
      this.canvas = document.getElementById(canvId);
      this.canvas.width = Game.projectionWidth;
      this.canvas.height = Game.projectionHeight;
      this.ctx = this.canvas.getContext("2d");
      let startPos = -this.sectionCount / 2;
      for (let i = 0; i < this.sectionCount; i++) {
        let projectionPosition = (startPos + i) * Game.sectWidth;
        let viewDistance = Math.hypot(projectionPosition,
          Game.playerDistance);
        this.steps.push(Math.asin(projectionPosition / viewDistance));
        this.dist.push(0); // * initialise distance array elements
      }
      this.halfStep = (Game.fov / this.sectionCount)*0.51;
      this.dist.push(0); // one extra element for max distance
    },
    frontAngle: function(dx, dy, rotation){
      let angle = Math.atan2(dy, dx) - rotation;
      // adjust angle to check it is front of player
      if(angle < -Math.PI){angle += Game.twoPi;}
      if(angle > Math.PI) {angle -= Game.twoPi;}
      return angle;
  }
};