function keyDown(e) {
    switch(e.keyCode){
        case 38: player.direction = 1; break; // up
        case 40: player.direction = -1; break; // down
        case 37: player.turn = -1; break; //left
        case 39: player.turn = 1; break; // right
        //case 32: player.weapons.activate(true); break; // space bar
        case 32: player.weapons.deploy(); break; // space bar
        case 88: player.changeWeapon(); break; // <x> key
    }
}
function keyUp(e) {
    switch(e.keyCode){
        case 38:
        case 40: player.direction = 0; break;
        case 37:
        case 39: player.turn = 0; break;
    }
}
function only360(angle) {
    angle %= Game.twoPi;
    while(angle < 0) {angle += Game.twoPi;}
    return angle;
}
// get nearest fraction of PI
function getNearestAngle(val){
    let l2 = [0,0.25,0.5,0.75,1,1.25,1.5,1.75,2];
    return l2.sort( (a, b) => Math.abs(val - a) - Math.abs(val - b) )[0];
}
// get index number of PI fraction
function getAngleIndex(angle) {
    const l1 = [0,0.25,0.5,0.75,1,1.25,1.5,1.75,2];
    return l1.indexOf(angle) % 8; // returns 0 to 7
}
function getAngleIndex4(val){
    const l1 = [0,0.5,1,1.5,2];
    let l2 = [0,0.5,1,1.5,2];
    let angle = l2.sort( (a, b) => Math.abs(val - a) - Math.abs(val - b) )[0];
    return l1.indexOf(angle) % 4; // returns 0 to 3
}
function findPointOnCircle(originX, originY, radius, angleRadians) {
    let circX = radius * Math.cos(angleRadians) + originX;
    let circY = radius * Math.sin(angleRadians) + originY;
    return {x: circX, y: circY};
}
function circIntersect(x1, y1, r1, x2, y2, r2){
    // strictly does not decide if circles intersect 
    // as one could be inside the other but works for this
    let dist = Math.hypot(x1-x2,y1-y2);
    return dist <= r1+r2;
}
// detectCollision used by player and NPCs
function detectCollision3(oldX, oldY, newX, newY, radius, rotation, direction){
    let retPos = {x: newX, y: newY}; // proposed new position
    let going = getAngleIndex4(rotation/Math.PI);
    if(direction < 0){ going = (going + 2) % 4;} // if reversing
    let angle = 0, hit = false, sHit = false, dim, s, pos;
    while(angle < Game.twoPi && !hit){
        pos = findPointOnCircle(newX, newY, radius, angle);
        if(Game.map.isWall(pos.x, pos.y)) {
            hit = true;}
        angle += Math.PI / 4;
    }
    if(!hit){
        for(let i= 0; i < sprites.length && !sHit; i++){
            s = sprites[i];
            if(!s.solid || !s.visible){continue;}
            dim = s.getDims().width;
            sHit = circIntersect(retPos.x, retPos.y, radius, s.x, s.y, dim/2);
            if(sHit){ // might be behind or to one side only
                angle = Math.atan2(s.y - retPos.y, s.x - retPos.x);
                if(getAngleIndex4(angle/Math.PI) !== going){
                    sHit = false; // brush past or going away
                }
            }
        }
    }
    if(!hit && !sHit){return retPos;}
    if(hit){
        pos = findPointOnCircle(newX, newY, radius, going * Math.PI/2);
        switch(going){
            case 0: // right
            case 2: // left
                if(Game.map.isWall(pos.x, pos.y)){retPos.x = oldX;}
                retPos.y = oldY;
                break;
            case 1: // down
            case 3: // up
                if(Game.map.isWall(pos.x, pos.y)){retPos.y = oldY;}
                retPos.x = oldX;
                break;
        }    
    } else {
        retPos.x = oldX;
        retPos.y = oldY;
    }
    return retPos;
}
function detectCollision2(curX, curY, newX, newY, radius){
    let angle = 0, pos;
    let retPos = {x: curX, y: curY};
    while (angle < Game.twoPi){
      pos = findPointOnCircle(newX, newY, radius, angle);
      if(Game.map.isWall(pos.x, pos.y)) {return retPos;}
      angle += Math.PI / 4;
    }
    for(let i = 0; i < sprites.length; i++){
      if(!sprites[i].solid){continue;}
      if(sprites[i].x === curX && sprites[i].y === curY){
          continue;
      }
      let dim = sprites[i].getDims();
      if(circIntersect(newX, newY, radius, sprites[i].x,
                            sprites[i].y, dim.width/2)){
        return retPos;
      }
    }
    return {x: newX, y: newY};
}
function createGrad(ctx, x0, y0, x1, y1, colour0, colour1, colour2){
    let grad = ctx.createLinearGradient(x0,y0,x1, y1);
    grad.addColorStop(0, colour0);
    if(arguments.length === 8){
        grad.addColorStop(0.5, colour1);
        grad.addColorStop(1, colour2);
    } else {
        grad.addColorStop(1, colour1);
    }
    return grad;
}
function setLevel(){
    Level.bgTopGrad = createGrad(camera.ctx, 0, 0, 0, Game.projectionHeight/2, "#FF7DE1", "#9B7DE1");
    Level.bgFloorGrad = createGrad(camera.ctx, 0, Game.projectionHeight/2, 0, Game.projectionHeight,  "#101010", "#303030", "#A0A0A0");
    Game.map = new FPSMap("devMap", "mapEntities", Level.grid);
    player.x = player.y = 200;
    for(let i = 0; i < Level.sprites.length; i++){
        let sp = Level.sprites[i];
        switch(sp.type){
            case 1:
                sprites.push(new Sprite(sp.image, sp.x, sp.y));
                break;
            case 2:
                sprites.push(new Sprite3D(sp.image, sp.x, sp.y, sp.fw, sp.rt, sp.fc));
                break;
            case 3:
                sprites.push(new NPC(sp.x, sp.y, sp.rt, sp.prog));
                break;
        }
    }
}
function setwalls() {
    // walls can be set at game level
    Game.walls.push(new Image());
    Game.walls[0].src = "images/stonewall.png";
    Game.walls.push(new Image());
    Game.walls[1].src = "images/stonebars.png";
    Game.walls.push(new Image());
    Game.walls[2].src = "images/stonecell.png";
}
function animateNPCs(){
    for(let i = 0; i < sprites.length; i++){
        if(sprites[i] instanceof NPC){
            sprites[i].animate();
        }
    }
}
function mapNPCs() {
    for(let i = 0; i < sprites.length; i++){
        if(sprites[i] instanceof NPC){
            if(!(sprites[i].state & NPC.states.dead)){
                Game.map.markNPC(sprites[i]);
            }
        }
    }
}
function renderScene() {
    player.animate();
    animateNPCs();
    camera.cast();
    mapNPCs();
    player.weapons.display(camera.ctx);
}
function initialise() {
    window.addEventListener("keydown", keyDown);
    window.addEventListener("keyup", keyUp);
    Game.playerDistance = (Game.projectionWidth / 2) / Math.tan(Game.fov / 2);
    camera.init("cvView");
    setwalls();
    setLevel();
    player.createWeapons();
    Game.timer = setInterval(renderScene, Game.speed);
}
initialise();