class FPSMap {
    constructor(mapCanvasId, thingsCanvasId, grid) {
        this.dMap = document.getElementById(mapCanvasId);
        this.tMap = document.getElementById(thingsCanvasId);
        this.mctx = this.dMap.getContext('2d');
        this.tctx = this.tMap.getContext('2d');
        this.scale = 16;
        this.grid = grid;
        let height = grid.length;
        let width = grid[0].length;
        this.dMap.height = this.tMap.height = height * this.scale;
        this.dMap.width = this.tMap.width = width * this.scale;
        this.tctx.clearRect(0,0,width,height);
        this.mctx.fillStyle = "white";
        this.mctx.fillRect(0,0,width,height);
        for(let h = 0; h < height; h++){
            for(let w = 0; w < width; w++){
                if(grid[h][w] !== 0){
                    this.mctx.fillStyle = "grey";
                    this.mctx.fillRect(w * this.scale, h * this.scale, this.scale, this.scale);
                }
            }
        }
    }
    showPlayer(player) {
        this.tctx.clearRect(0,0,this.tMap.width,this.tMap.height);
        this.tctx.beginPath();
        let x = player.x * this.scale / Game.gameScale;
        let y = player.y * this.scale / Game.gameScale;
        this.tctx.arc(x, y, player.mapRadius, 0, Game.twoPi);
        this.tctx.fillStyle = "blue";
        this.tctx.fill();
        this.tctx.closePath();
        this.tctx.beginPath();
        this.tctx.moveTo(x,y);
        x += Math.cos(player.rotation) * player.mapRadius * 3;
        y += Math.sin(player.rotation) * player.mapRadius * 3;
        this.tctx.lineTo(x,y);
        this.tctx.strokeStyle = "red";
        this.tctx.stroke();
        this.tctx.closePath();
    }
    isWall(x,y){
        y = Math.floor(y/Game.gameScale);
        x = Math.floor(x/Game.gameScale);
        return this.grid[y][x] !== 0;
    }
    getGrid(x, y){
        return this.grid[y][x];
    }
    getWidth() {
        return this.grid[0].length;
    }
    getHeight() {
        return this.grid.length;
    }
    drawRay(player, pt){
        let x = player.x * this.scale / Game.gameScale;
        let y = player.y * this.scale / Game.gameScale;
        this.tctx.beginPath();
        this.tctx.moveTo(x,y);
        this.tctx.lineTo(pt.x * this.scale,pt.y * this.scale);
        this.tctx.strokeStyle = "aqua";
        this.tctx.lineWidth = 0.5;
        this.tctx.stroke();
        this.tctx.closePath();    
    }
    markNPC(npc){
        this.tctx.beginPath();
        let x = npc.x * this.scale / Game.gameScale;
        let y = npc.y * this.scale / Game.gameScale;
        this.tctx.arc(x, y, npc.mapRadius, 0, Game.twoPi);
        this.tctx.fillStyle = "red";
        this.tctx.fill();
        this.tctx.closePath();
    }
}